"""Stationary combustion models"""
# pylint: disable=no-name-in-module
import logging
from atomic6ghg.formulas import Formula

from atomic6ghg.factors import heat_content_factors, stationary_combustion_emission_factors, refrigerants_gwp_factors

logger = logging.getLogger(__name__)


class StationaryCombustion(Formula):
    """ Stationary combustion includes fuels that are burned at stationary facilities. """

    fossil_fuels = ['anthraciteCoal', 'bituminousCoal', 'subBituminousCoal', 'ligniteCoal', 'naturalGas',
                    'distillateFuelOilNo2', 'residualFuelOilNo6', 'kerosene', 'liquefiedPetroleumGases']
    non_fossil_fuels = ['woodAndWoodResiduals', 'landfillGas']
    all_fuels = fossil_fuels + non_fossil_fuels

    common_units = {'anthraciteCoal': 'shortTon', 'bituminousCoal': 'shortTon',
                    'distillateFuelOilNo2': 'gallons', 'kerosene': 'gallons', 'landfillGas': 'scf',
                    'ligniteCoal': 'shortTon', 'liquefiedPetroleumGases': 'gallons', 'naturalGas': 'scf',
                    'residualFuelOilNo6': 'gallons', 'subBituminousCoal': 'shortTon',
                    'woodAndWoodResiduals': 'shortTon'}

    def __init__(self, wks_data=None):
        super().__init__(wks_data=wks_data)
        self.recalc(self.wks_data)

    def recalc(self, wks_data: dict) -> dict:
        """Execute recalc procedure for StationaryCombustion"""
        self.wks_data = wks_data

        self.make_total_combustion()
        self.make_emissions()
        self.make_co2_equivalent_emissions()
        self.make_biomass_co2_equivalent_emissions()

        # Add user data to _output
        self._output['stationarySourceFuelConsumption'] = self.wks_data.get('stationarySourceFuelConsumption', [])

        return self.to_dict()

    def make_total_combustion(self):
        """Calculate total combustion for all input rows"""
        total_combustion = {fuel: 0. for fuel in self.all_fuels}
        for row in self.wks_data.get('stationarySourceFuelConsumption', []):
            fuel = row['fuelCombusted']
            quantity_combusted = row['quantityCombusted']
            units = row['units']
            if not fuel or not quantity_combusted or not units:
                continue
            total_combustion[fuel] += heat_content_factors[fuel][units] * quantity_combusted

        # Flattening the dictionary
        total_combustion = [{'fuelType': fuel, 'quantityCombusted': quantity_combusted,
                             'units': self.common_units[fuel]} for fuel, quantity_combusted in total_combustion.items()]

        self._output['totalStationarySourceCombustion'] = total_combustion

    def make_emissions(self):
        """Calculate emissions for all fuels burned"""
        emissions = {fuel: {'CO2': 0., 'CH4': 0., 'N2O': 0.} for fuel in self.all_fuels}
        emissions['totalFossilFuelEmissions'] = {'CO2': 0., 'CH4': 0., 'N2O': 0.}
        emissions['totalNonFossilFuelEmissions'] = {'CO2': 0., 'CH4': 0., 'N2O': 0.}

        for row in self._output['totalStationarySourceCombustion']:
            fuel = row['fuelType']
            total_fuel_combustion = row['quantityCombusted']
            if not total_fuel_combustion:
                continue
            emissions[fuel]['CO2'] = StationaryCombustion.calculate_co2_emissions(fuel, total_fuel_combustion)
            emissions[fuel]['CH4'] = StationaryCombustion.calculate_ch4_emissions(fuel, total_fuel_combustion)
            emissions[fuel]['N2O'] = StationaryCombustion.calculate_n2o_emissions(fuel, total_fuel_combustion)

            # There should not be any fuels not in either list at this point
            if fuel in self.fossil_fuels:
                key = 'totalFossilFuelEmissions'
            else:
                key = 'totalNonFossilFuelEmissions'
            emissions[key]['CO2'] += emissions[fuel]['CO2']
            emissions[key]['CH4'] += emissions[fuel]['CH4']
            emissions[key]['N2O'] += emissions[fuel]['N2O']

        emissions['totalEmissionsForAllFuels'] = {gas: emissions['totalFossilFuelEmissions'][gas] +
                                                       emissions['totalNonFossilFuelEmissions'][gas]
                                                  for gas in emissions['totalFossilFuelEmissions']}

        self._total_emissions = {'totalFossilFuelEmissions': emissions['totalFossilFuelEmissions'],
                                 'totalNonFossilFuelEmissions': emissions['totalNonFossilFuelEmissions'],
                                 'totalEmissionsForAllFuels': emissions['totalEmissionsForAllFuels'],
                                 }

        # Flattening the dictionary
        emissions = [{'fuelType': fuel, **emissions[fuel]} for fuel in emissions]
        self._output['totalGhgEmissionsFromStationarySourceFuelCombustion'] = emissions

    @staticmethod
    def calculate_co2_emissions(fuel, total_fuel_combustion):
        """Calculate CO2 emissions for a fuel"""
        return stationary_combustion_emission_factors[fuel]['CO2 Factor (kg / Unit)'] * total_fuel_combustion

    @staticmethod
    def calculate_ch4_emissions(fuel, total_fuel_combustion):
        """Calculate CH4 emissions for a fuel"""
        return stationary_combustion_emission_factors[fuel]['CH4 Factor (g / unit)'] * total_fuel_combustion

    @staticmethod
    def calculate_n2o_emissions(fuel, total_fuel_combustion):
        """Calculate N2O emissions for a fuel"""
        return stationary_combustion_emission_factors[fuel]['N2O Factor (g / unit)'] * total_fuel_combustion

    def make_co2_equivalent_emissions(self):
        """ Calculate co2 equivalent emissions """
        # Units of CH4 and N2O emissions are in grams and need to be converted to kilograms
        conversions = {'CO2': 1., 'CH4': 1 / 1000., 'N2O': 1 / 1000.}
        total_fossil = sum([self._total_emissions['totalFossilFuelEmissions'][gas] *
                            refrigerants_gwp_factors[gas.lower()] * conversions[gas]
                            for gas in self._total_emissions['totalFossilFuelEmissions']])
        total_nonfossil = sum([self._total_emissions['totalNonFossilFuelEmissions'][gas] *
                               refrigerants_gwp_factors[gas.lower()] * conversions[gas]
                               for gas in self._total_emissions['totalNonFossilFuelEmissions'] if gas != 'CO2'])
        total = (total_fossil + total_nonfossil) / 1000.

        self._output['totalCO2EquivalentEmissions'] = total

    def make_biomass_co2_equivalent_emissions(self):
        """Calculate non-fossil-fuel emissions"""
        total = self._total_emissions['totalNonFossilFuelEmissions']['CO2'] / 1000.

        self._output['totalBiomassEquivalentEmissions'] = total



""" Module to wrap factors in class """
import json
import pkgutil


class FireSuppressionFactors:
    """ Wrapper class for fire_suppression_leak_rate_factors.json """
    def __init__(self):
        self.factors = json.loads(pkgutil.get_data('atomic6ghg.factors',
            'source_data/fire_suppression_leak_rates_factors.json'))

    def __getitem__(self, item):
        return self.factors.get(item)



""" Module to wrap factors in class """
import json
import pkgutil
from atomic6ghg import YearValueException, YearMapException

class MobileCombustionCh4AndN2oEmissionFactors:
    """ Wrapper class for mobile_combustion_emission_factors.json """
    def __init__(self):
        self.factors = json.loads(pkgutil.get_data('atomic6ghg.factors',
            'source_data/mobile_combustion_ch4_and_n2o_emission_factors.json'))

        self._factors = {}
        self.make_factors()

    def __getitem__(self, vehicle_type):
        return self._factors.get(vehicle_type)

    class YearHandler:
        """ Class to handle user input years to map the correct factors for a given vehicle type and fuel type """
        def __init__(self, factors_by_year):
            self.factors_by_year = {}
            self._rekey_years(factors_by_year)

            self.year_boundaries = []
            if len(self.factors_by_year) > 1:
                self.get_year_boundaries()

        def _rekey_years(self, factors_by_year):
            """ Turn string keys into integers """

            # Assume if length is one then the only key is ''
            if len(factors_by_year) == 1:
                self.factors_by_year = factors_by_year
            else:
                self.factors_by_year = {int(year): factor for year, factor in factors_by_year.items()}

        def get_year_boundaries(self):
            """ Get the minimum and maximum years that have factors """
            years = sorted([int(year) for year in self.factors_by_year])
            self.year_boundaries = [years[0], years[-1]]

        @staticmethod
        def _int_year(year):
            try:
                if isinstance(year, str):
                    assert year.isdigit()
                elif isinstance(year, float):
                    assert year.is_integer()
                else:
                    assert isinstance(year, int)
                return int(year)
            except AssertionError as e:
                raise YearValueException from e

        def __getitem__(self, year):
            # If there is only one key then the fuel type has no time dependency and has key 'default'.
            # This also ignores year, which the user can input and may be passed along to the factors.
            if len(self.factors_by_year) == 1:
                return self.factors_by_year['default']
            year = self._int_year(year)
            try:
                return self.factors_by_year[year]
            except KeyError:
                pass
            if year < self.year_boundaries[0]:
                ret = self[self.year_boundaries[0]]
            elif year > self.year_boundaries[1]:
                ret = self[self.year_boundaries[1]]
            else:
                raise YearMapException
            return ret

        def __repr__(self):
            return repr(self.factors_by_year)

        def __iter__(self):
            return iter(self.factors_by_year)

    def __repr__(self):
        return repr(self._factors)

    def __iter__(self):
        return iter(self._factors)

    def make_factors(self):
        """ Wrap year level of factors json in class YearHandler to handle user input for years of mobile vehicles"""
        for vehicle_type in self.factors:
            self._factors[vehicle_type] = {}
            for fuel_type in self.factors[vehicle_type]:
                self._factors[vehicle_type][fuel_type] = self.YearHandler(self.factors[vehicle_type][fuel_type])



""" Waste Gases models """
# pylint: disable=no-name-in-module
import logging

from atomic6ghg.formulas import Formula, null_replacer
from atomic6ghg.factors import molecular_weights_factors, unit_conversions_factors

logger = logging.getLogger(__name__)


class WasteGases(Formula):
    """ Calculate emission from combustion of waste gases """

    default_gas_total_number_of_moles_per_unit_volume = 0.00255
    default_oxidation_factor = 100.

    def __init__(self, wks_data=None):
        super().__init__(wks_data=wks_data)
        self.gas_total_number_of_moles_per_unit_volume = None
        self.oxidation_factor = None
        self.recalc(self.wks_data)

    def recalc(self, wks_data):
        """ Execute recalc procedure for waste gases """
        self.wks_data = wks_data

        self.gas_total_number_of_moles_per_unit_volume = self.wks_data.get("gasTotalNumberOfMolesPerUnitVolume") \
            if self.wks_data.get("gasTotalNumberOfMolesPerUnitVolume") \
            else self.default_gas_total_number_of_moles_per_unit_volume
        self.oxidation_factor = self.wks_data.get("oxidationFactor") \
            if self.wks_data.get("oxidationFactor") else self.default_oxidation_factor

        self.make_emission_factor_for_gas_waste_stream()
        self.make_total_all_components()
        self.make_co2_equivalent_emissions()

        self._output["wasteStreamGasCombusted"] = self.wks_data['wasteStreamGasCombusted']
        self._output["gasTotalNumberOfMolesPerUnitVolume"] = self.gas_total_number_of_moles_per_unit_volume
        self._output["oxidationFactor"] = self.oxidation_factor

        return self.to_dict()

    def make_emission_factor_for_gas_waste_stream(self):
        """ Calculate emission for each inputs in factor for gases waste stream table """
        waste_gasses = []
        self._total_emissions = {'totalCarbonContent': 0., 'totalMoles': 0., 'totalMolarFraction': 0.}
        gas_total_number_of_moles_per_unit_volume = self.gas_total_number_of_moles_per_unit_volume
        for row in self.wks_data.get('emissionFactorForGasWasteStream', []):
            component = row.get('component')
            if not component:
                waste_gasses.append(row)
                continue

            molar_fraction = null_replacer(row['molarFraction'])
            if component != 'Other non-carbon':
                chemical_formula = molecular_weights_factors[component]['chemicalFormula']
                molecular_weight = molecular_weights_factors[component]['molecularWeight']
                percent_carbon = molecular_weights_factors[component]['percentCarbon']
            else:
                chemical_formula = None
                molecular_weight = 0
                percent_carbon = 0

            total_moles = WasteGases.calculate_total_moles(molar_fraction, gas_total_number_of_moles_per_unit_volume)
            carbon_content = WasteGases.calculate_carbon_content(total_moles, molecular_weight, percent_carbon)

            self._total_emissions['totalMoles'] += total_moles
            self._total_emissions['totalMolarFraction'] += molar_fraction
            self._total_emissions['totalCarbonContent'] += carbon_content

            calculated_row = {'component': component,
                              'chemicalFormula': chemical_formula,
                              'molarFraction': molar_fraction,
                              'totalMoles': total_moles,
                              'molecularWeight': molecular_weight,
                              'percentCarbon': percent_carbon,
                              'carbonContent': carbon_content
                              }
            waste_gasses.append(calculated_row)

        self._output['emissionFactorForGasWasteStream'] = waste_gasses

    def make_total_all_components(self):
        """ Make total row """
        total_all_components = {'molarFraction': self._total_emissions['totalMolarFraction'],
                                'totalMoles': self._total_emissions['totalMoles'],
                                'carbonContent': self._total_emissions['totalCarbonContent']
                                }

        self._output['totalForAllComponents'] = total_all_components

    @staticmethod
    def calculate_total_moles(molar_fraction, gas_total_number_of_moles_per_unit_volume):
        """ Calculates total moles for given component"""
        ret = molar_fraction / 100 * gas_total_number_of_moles_per_unit_volume
        ret = max(0., ret)
        return ret

    @staticmethod
    def calculate_carbon_content(total_moles, molecular_weight, percent_carbon):
        """Calculates carbon content for given component"""
        ret = total_moles * molecular_weight * percent_carbon
        ret = max(0., ret)
        return ret

    def make_co2_equivalent_emissions(self):
        """ Calculates co2 equivalent for all components """
        waste_stream_gas_combusted = self.wks_data.get('wasteStreamGasCombusted')
        if waste_stream_gas_combusted:
            total = self._total_emissions['totalCarbonContent'] * self.oxidation_factor / 100. \
                    * (44.0098 / 12.001) \
                    * waste_stream_gas_combusted * unit_conversions_factors['pounds']['kilogram'] \
                    / 1000.
        else:
            # TODO: Raise error in logging message
            total = 0.

        self._output['totalCO2EquivalentEmissions'] = total



""" Commuting models """
# pylint: disable=no-name-in-module
import logging

from atomic6ghg.formulas import Formula, null_replacer
from atomic6ghg.factors import business_travel_factors, refrigerants_gwp_factors

logger = logging.getLogger(__name__)


class Commuting(Formula):
    """ Calculate emissions from commuting equipment """

    personal_vehicles = ['passengerCars', 'lightDutyTruck', 'motorcycle']
    public_transit = ['intercityRailNortheastCorridor', 'intercityRailOtherRoutes', 'intercityRailNationalAverage',
                      'commuterRail', 'transitRail', 'bus']
    all_vehicles = personal_vehicles + public_transit

    def __init__(self, wks_data=None):
        super().__init__(wks_data=wks_data)
        self._total_emissions_by_vehicle = {}
        self._total_emissions_by_transit = {}
        self.recalc(self.wks_data)

    def recalc(self, wks_data: dict) -> dict:
        """Execute recalc procedure for Commuting"""
        self.wks_data = wks_data

        self.make_personal_vehicle()
        self.make_total_for_all_personal_vehicle()
        self.make_public_transportation()
        self.make_total_for_all_public_transportation()
        self.make_emissions_by_commuting_type()
        self.make_co2_equivalent_emissions()

        return self.to_dict()

    def make_personal_vehicle(self):
        """Calculate CO2 equivalent emissions for each input row in Personal Vehicle table"""
        personal_vehicle = []
        self._total_emissions = {
            'personalVehicle': {
                'CO2': 0.,
                'CH4': 0.,
                'N2O': 0.
            }
        }

        self._total_emissions_by_vehicle = {vehicle: {
            'CO2': 0.,
            'CH4': 0.,
            'N2O': 0.
        } for vehicle in self.personal_vehicles}

        for row in self.wks_data.get('personalVehicle', []):
            vehicle_type = row['vehicleType']
            vehicle_miles = row['vehicleMiles']
            if not vehicle_type or not vehicle_miles:
                personal_vehicle.append(row)
                continue
            if vehicle_type:
                travel_factor_co2 = business_travel_factors[vehicle_type]['co2']
                travel_factor_ch4 = business_travel_factors[vehicle_type]['ch4']
                travel_factor_n2o = business_travel_factors[vehicle_type]['n2o']
            else:
                travel_factor_co2, travel_factor_ch4, travel_factor_n2o = 0., 0., 0.

            vehicle_miles = null_replacer(row['vehicleMiles'])

            co2_emissions = Commuting.calculate_transportation_emissions(travel_factor_co2, vehicle_miles)
            ch4_emissions = Commuting.calculate_transportation_emissions(travel_factor_ch4, vehicle_miles)
            n2o_emissions = Commuting.calculate_transportation_emissions(travel_factor_n2o, vehicle_miles)

            self._total_emissions['personalVehicle']['CO2'] += co2_emissions
            self._total_emissions['personalVehicle']['CH4'] += ch4_emissions
            self._total_emissions['personalVehicle']['N2O'] += n2o_emissions
            self._total_emissions_by_vehicle[vehicle_type]['CO2'] += co2_emissions
            self._total_emissions_by_vehicle[vehicle_type]['CH4'] += ch4_emissions
            self._total_emissions_by_vehicle[vehicle_type]['N2O'] += n2o_emissions

            calculated_row = {'sourceId': row.get('sourceId'), 'sourceDescription': row.get('sourceDescription'),
                              'vehicleType': vehicle_type, 'vehicleMiles': vehicle_miles, 'CO2': co2_emissions,
                              'CH4': ch4_emissions, 'N2O': n2o_emissions}

            personal_vehicle.append(calculated_row)

        self._output['personalVehicle'] = personal_vehicle

    def make_total_for_all_personal_vehicle(self):
        """ Make total row """
        total_for_all_personal_vehicle = {
                     'CO2': self._total_emissions['personalVehicle']['CO2'],
                     'CH4': self._total_emissions['personalVehicle']['CH4'],
                     'N2O': self._total_emissions['personalVehicle']['N2O']}

        self._output['totalForAllPersonalVehicleEmployeeCommuting'] = total_for_all_personal_vehicle

    def make_public_transportation(self):
        """Calculate emissions for each input row in Public Transportation table"""
        public_transportation = []
        self._total_emissions.update({
            'publicTransportation': {
                'CO2': 0.,
                'CH4': 0.,
                'N2O': 0.
            }
        })

        self._total_emissions_by_transit = {transit: {
            'CO2': 0.,
            'CH4': 0.,
            'N2O': 0.
        } for transit in self.public_transit}

        for row in self.wks_data.get('publicTransportation', []):
            transport = row['transportType']
            passenger_miles = row['passengerMiles']
            if not transport or not passenger_miles:
                public_transportation.append(row)
                continue
            if transport:
                travel_factor_co2 = business_travel_factors[transport]['co2']
                travel_factor_ch4 = business_travel_factors[transport]['ch4']
                travel_factor_n2o = business_travel_factors[transport]['n2o']
            else:
                travel_factor_co2, travel_factor_ch4, travel_factor_n2o = 0., 0., 0.

            passenger_miles = null_replacer(row['passengerMiles'])

            co2_emissions = Commuting.calculate_transportation_emissions(travel_factor_co2, passenger_miles)
            ch4_emissions = Commuting.calculate_transportation_emissions(travel_factor_ch4, passenger_miles)
            n2o_emissions = Commuting.calculate_transportation_emissions(travel_factor_n2o, passenger_miles)

            self._total_emissions['publicTransportation']['CO2'] += co2_emissions
            self._total_emissions['publicTransportation']['CH4'] += ch4_emissions
            self._total_emissions['publicTransportation']['N2O'] += n2o_emissions
            self._total_emissions_by_transit[transport]['CO2'] += co2_emissions
            self._total_emissions_by_transit[transport]['CH4'] += ch4_emissions
            self._total_emissions_by_transit[transport]['N2O'] += n2o_emissions

            calculated_row = {'sourceId': row['sourceId'], 'sourceDescription': row['sourceDescription'],
                              'transportType': transport, 'passengerMiles': passenger_miles, 'CO2': co2_emissions,
                              'CH4': ch4_emissions, 'N2O': n2o_emissions}

            public_transportation.append(calculated_row)

        self._output['publicTransportation'] = public_transportation

    def make_total_for_all_public_transportation(self):
        """ Make total row """
        total_for_all_public_transportation = {
                     'CO2': self._total_emissions['publicTransportation']['CO2'],
                     'CH4': self._total_emissions['publicTransportation']['CH4'],
                     'N2O': self._total_emissions['publicTransportation']['N2O']}

        self._output['totalForAllPublicTransportationEmployeeCommuting'] = total_for_all_public_transportation

    def make_emissions_by_commuting_type(self):
        """Calculate CO2 equivalent emissions for each input row"""
        total_commuting_type_co2 = []
        self._total_emissions['total'] = {'CO2': 0., 'CH4': 0., 'N2O': 0.}
        self._total_emissions.update(self._total_emissions_by_vehicle)
        self._total_emissions.update(self._total_emissions_by_transit)
        for transport_type in self.all_vehicles:

            self._total_emissions['total']['CO2'] += self._total_emissions[transport_type]['CO2']
            self._total_emissions['total']['CH4'] += self._total_emissions[transport_type]['CH4']
            self._total_emissions['total']['N2O'] += self._total_emissions[transport_type]['N2O']

            calculated_row = {'transportType': transport_type, 'CO2': self._total_emissions[transport_type]['CO2'],
                              'CH4': self._total_emissions[transport_type]['CH4'],
                              'N2O': self._total_emissions[transport_type]['N2O']}
            total_commuting_type_co2.append(calculated_row)
        self._output['totalCO2EmissionsByCommutingType'] = total_commuting_type_co2

    @staticmethod
    def calculate_transportation_emissions(travel_factor, vehicle_miles):
        """Calculate emissions for a gas given travel factor and vehicle miles inputs"""
        ret = travel_factor * vehicle_miles
        # Handle potential negative values
        ret = max(0., ret)
        return ret

    def make_co2_equivalent_emissions(self):
        """Calculate co2 equivalent emissions """
        total = (self._total_emissions['total']['CO2'] +
                 (self._total_emissions['total']['CH4'] / 1000 * refrigerants_gwp_factors['ch4']) +
                 (self._total_emissions['total']['N2O'] / 1000 * refrigerants_gwp_factors['n2o'])) / 1000
        self._output['totalCO2EquivalentEmissions'] = total



""" Custom exceptions """


class YearMapException(Exception):
    """ This exception is raised if a year can't be mapped to factors """


class YearValueException(Exception):
    """ This exception is raised if a year cannot be typed as an int """



""""Console API for atomic6ghg"""
# __main__.py


def main() -> None:
    """Reflect a description of the current package"""
    print("Package under construction.")


if __name__ == "__main__":
    main()


""" Mobile Sources models """
# pylint: disable=no-name-in-module
import logging

from atomic6ghg.formulas import Formula
from atomic6ghg.factors import mobile_combustion_co2_emission_factors, mobile_combustion_ch4_and_n2o_emission_factors, \
    refrigerants_gwp_factors
from atomic6ghg import YearMapException, YearValueException

logger = logging.getLogger(__name__)


class MobileSources(Formula):
    """ Calculate emissions from mobile source vehicles """

    co2_fuels_units = {'gasoline': 'gallons', 'diesel': 'gallons', 'residualFuelOil': 'gallons',
                       'aviationGasoline': 'gallons', 'jetFuel': 'gallons', 'lpg': 'gallons',
                       'ethanol': 'gallons', 'biodiesel': 'gallons', 'lng': 'gallons', 'cng': 'scf'}
    co2_fuel_map = {'gasoline2Stroke': 'gasoline', 'gasoline4Stroke': 'gasoline', 'methanol': 'ethanol'}
    non_biomass_co2_fuel_map = {'ethanol': 'gasoline', 'biodiesel': 'diesel'}
    road_vehicles = ['heavyDutyVehicles', 'lightDutyTrucks', 'motorcycles', 'passengerCars', 'buses', 'heavyDutyTrucks',
                     'lightDutyCars', 'mediumAndHeavyDutyVehicles', 'mediumDutyTrucks']
    nonroad_vehicles = ['agriculturalEquipment', 'agriculturalOffroadTrucks', 'aircraft', 'airportEquipment',
                        'constructionMiningEquipment', 'constructionMiningOffroadTrucks',
                        'industrialCommercialEquipment', 'lawnAndGardenEquipment', 'locomotives', 'loggingEquipment',
                        'railroadEquipment', 'recreationalEquipment', 'shipsAndBoats']

    def __init__(self, wks_data=None):
        super().__init__(wks_data=wks_data)

        self.biodiesel_percent = 20
        self.ethanol_percent = 80

        self.total_fuel_usage_and_co2_emissions = {}
        self.total_useage_and_ch4_and_n2o_emissions = {}

        self.recalc(self.wks_data)

    def recalc(self, wks_data: dict) -> dict:
        """ Execute recalc procedure for MobileSources """
        self.wks_data = wks_data

        self._total_emissions = {'CO2': 0., 'CH4': 0., 'N2O': 0.}

        self.total_fuel_usage_and_co2_emissions = {fuel_type: {'fuelUsage': 0, 'CO2': 0.}
                                                   for fuel_type in self.co2_fuels_units}

        self.total_useage_and_ch4_and_n2o_emissions = \
            {vehicle_type: {fuel_type: {
                mobile_combustion_ch4_and_n2o_emission_factors[vehicle_type][fuel_type][year]['year_display']:
                    {'N2O': 0., 'CH4': 0., 'mileage': 0., 'fuelUsage': 0.}
                for year in mobile_combustion_ch4_and_n2o_emission_factors[vehicle_type][fuel_type]}
                for fuel_type in mobile_combustion_ch4_and_n2o_emission_factors[vehicle_type]}
                for vehicle_type in mobile_combustion_ch4_and_n2o_emission_factors
            }

        # If these keys aren't in wks_data then default to original wks_data values
        self.biodiesel_percent = self.wks_data.get('biodieselPercent', 20)
        self.ethanol_percent = self.wks_data.get('ethanolPercent', 80)

        self.tabulate_subtable_data()

        self.make_total_mobile_sources_fuel_usage_and_co2_emissions()
        self.make_total_organization_wide_on_road_gasoline_mobile_source_mileage_and_emissions()
        self.make_total_organization_wide_on_road_non_gasoline_mobile_source_mileage_and_emissions()
        self.make_total_organization_wide_non_road_mobile_source_fuel_usage_and_emissions()
        self.make_total_co2_equivalent_emissions()
        self.make_total_biomass_co2_equivalent_emissions()

        # Add user data to _output
        self._output['mobileSourcesFuelConsumption'] = self.wks_data.get('mobileSourcesFuelConsumption', [])
        self._output['biodieselPercent'] = self.biodiesel_percent
        self._output['ethanolPercent'] = self.ethanol_percent

        return self.to_dict()

    def tabulate_subtable_data(self):
        """ Loop over user input data and make all sub-tables. """

        for row in self.wks_data.get('mobileSourcesFuelConsumption', []):
            vehicle_type = row['vehicleType']
            fuel_usage = row['fuelUsage']
            miles_traveled = row['milesTraveled']
            fuel_type = row['fuelType']
            vehicle_year = row['vehicleYear']
            if not fuel_type:
                continue
            if fuel_usage:
                self.tabulate_total_mobile_sources_fuel_usage_and_co2_emissions(fuel_type, fuel_usage)
            if vehicle_type:
                self.tabulate_mobile_source_use_and_emissions_by_vehicle_type(vehicle_type, fuel_type, vehicle_year,
                    fuel_usage, miles_traveled)

    def tabulate_total_mobile_sources_fuel_usage_and_co2_emissions(self, fuel_type, fuel_usage):
        """Calculate CO2 equivalent emissions for each input fuel_type and fuel_usage"""

        co2_fuel_type = self.co2_fuel_map.get(fuel_type, fuel_type)
        co2_fuel_usage = self.get_fuel_usage(co2_fuel_type, fuel_usage)
        non_biomass_co2_fuel_type = self.non_biomass_co2_fuel_map.get(co2_fuel_type, co2_fuel_type)

        co2_equivalent_emissions = MobileSources.calculate_co2_emissions(non_biomass_co2_fuel_type, co2_fuel_usage)

        self.total_fuel_usage_and_co2_emissions[co2_fuel_type]['fuelUsage'] += fuel_usage
        self.total_fuel_usage_and_co2_emissions[co2_fuel_type]['CO2'] += co2_equivalent_emissions

        self._total_emissions['CO2'] += co2_equivalent_emissions

    def get_fuel_usage(self, fuel_type, fuel_usage):
        """ Get non biomass fuel usage for each fuel type """
        if fuel_type == 'ethanol':
            ret = fuel_usage * ((100 - self.ethanol_percent) / 100)
        elif fuel_type == 'biodiesel':
            ret = fuel_usage * ((100 - self.biodiesel_percent) / 100)
        else:
            ret = fuel_usage
        return ret

    # pylint: disable=too-many-arguments
    def tabulate_mobile_source_use_and_emissions_by_vehicle_type(self, vehicle_type, fuel_type, vehicle_year,
                                                                 fuel_usage, miles_traveled):
        """ Calculate total fuel usage, total mileage, and CH4 and N2O emissions for vehicles """

        if vehicle_type in self.road_vehicles:
            usage = miles_traveled
        else:
            usage = fuel_usage

        try:
            ch4_emissions = MobileSources.calculate_ch4_emissions(vehicle_type, fuel_type, usage, vehicle_year)
        except (YearMapException, YearValueException) as e:
            logger.error(e)
            ch4_emissions = 0.

        try:
            n2o_emissions = MobileSources.calculate_n2o_emissions(vehicle_type, fuel_type, usage, vehicle_year)
        except (YearMapException, YearValueException) as e:
            logger.error(e)
            n2o_emissions = 0.

        year_display = \
            mobile_combustion_ch4_and_n2o_emission_factors[vehicle_type][fuel_type][vehicle_year]['year_display']
        self.total_useage_and_ch4_and_n2o_emissions[vehicle_type][fuel_type][year_display]['CH4'] += ch4_emissions
        self.total_useage_and_ch4_and_n2o_emissions[vehicle_type][fuel_type][year_display]['N2O'] += n2o_emissions
        self.total_useage_and_ch4_and_n2o_emissions[vehicle_type][fuel_type][year_display]['mileage'] += miles_traveled
        self.total_useage_and_ch4_and_n2o_emissions[vehicle_type][fuel_type][year_display]['fuelUsage'] += fuel_usage

        self._total_emissions['CH4'] += ch4_emissions
        self._total_emissions['N2O'] += n2o_emissions

    def make_total_mobile_sources_fuel_usage_and_co2_emissions(self):
        """ Format total_fuel_usage_and_co2_emissions data for schema """
        total_fuel_usage_and_co2_emissions = \
            [{'fuelType': fuel_type,
              'fuelUsage': self.total_fuel_usage_and_co2_emissions[fuel_type]['fuelUsage'],
              'units': self.co2_fuels_units[fuel_type],
              'CO2': self.total_fuel_usage_and_co2_emissions[fuel_type]['CO2']}
             for fuel_type in self.total_fuel_usage_and_co2_emissions]

        self._output['totalMobileSourcesFuelUsageAndCO2Emissions'] = total_fuel_usage_and_co2_emissions

    def make_total_organization_wide_on_road_gasoline_mobile_source_mileage_and_emissions(self):
        """ Select gasoline on road vehicle data and format total_useage_and_ch4_and_n2o_emissions for schema """

        total_on_road_gasoline_mileage_and_emissions = \
            [{'vehicleType': vehicle_type,
              'emissionByYear': [
                  {'vehicleYear': year,
                   'mileage': self.total_useage_and_ch4_and_n2o_emissions[vehicle_type][fuel_type][year]['mileage'],
                   'CH4': self.total_useage_and_ch4_and_n2o_emissions[vehicle_type][fuel_type][year]['CH4'],
                   'N2O': self.total_useage_and_ch4_and_n2o_emissions[vehicle_type][fuel_type][year]['N2O']}
                  for year in self.total_useage_and_ch4_and_n2o_emissions[vehicle_type][fuel_type]]}
             for vehicle_type in self.total_useage_and_ch4_and_n2o_emissions
             for fuel_type in self.total_useage_and_ch4_and_n2o_emissions[vehicle_type]
             if vehicle_type in self.road_vehicles and fuel_type == 'gasoline']

        self._output['totalOrganizationWideOnRoadGasolineMobileSourceMileageAndEmissions'] = \
            total_on_road_gasoline_mileage_and_emissions

    def make_total_organization_wide_on_road_non_gasoline_mobile_source_mileage_and_emissions(self):
        """ Select non-gasoline on road vehicle data and format total_useage_and_ch4_and_n2o_emissions for schema """

        total_on_road_non_gasoline_mileage_and_emissions = \
            [{'vehicleType': vehicle_type,
              'fuelType': fuel_type,
              'emissionByYear': [
                  {'vehicleYear': year,
                   'mileage': self.total_useage_and_ch4_and_n2o_emissions[vehicle_type][fuel_type][year]['mileage'],
                   'CH4': self.total_useage_and_ch4_and_n2o_emissions[vehicle_type][fuel_type][year]['CH4'],
                   'N2O': self.total_useage_and_ch4_and_n2o_emissions[vehicle_type][fuel_type][year]['N2O']}
                  for year in self.total_useage_and_ch4_and_n2o_emissions[vehicle_type][fuel_type]]}
             for vehicle_type in self.total_useage_and_ch4_and_n2o_emissions
             for fuel_type in self.total_useage_and_ch4_and_n2o_emissions[vehicle_type]
             if vehicle_type in self.road_vehicles and fuel_type != 'gasoline']

        self._output['totalOrganizationWideOnRoadNonGasolineMobileSourceMileageAndEmissions'] = \
            total_on_road_non_gasoline_mileage_and_emissions

    def make_total_organization_wide_non_road_mobile_source_fuel_usage_and_emissions(self):
        """ Select non-road vehicle data and format total_useage_and_ch4_and_n2o_emissions for schema """

        total_on_road_non_gasoline_mileage_and_emissions = \
            [{'vehicleType': vehicle_type,
              'emissionByFuelType': [
                  {'fuelType': fuel_type,
                   'fuelUsage': self.total_useage_and_ch4_and_n2o_emissions[vehicle_type][fuel_type]['']['fuelUsage'],
                   'CH4': self.total_useage_and_ch4_and_n2o_emissions[vehicle_type][fuel_type]['']['CH4'],
                   'N2O': self.total_useage_and_ch4_and_n2o_emissions[vehicle_type][fuel_type]['']['N2O']}
                  for fuel_type in self.total_useage_and_ch4_and_n2o_emissions[vehicle_type]]}
             for vehicle_type in self.total_useage_and_ch4_and_n2o_emissions
             if vehicle_type in self.nonroad_vehicles]

        self._output['totalOrganizationWideNonRoadMobileSourceFuelUsageAndEmissions'] = \
            total_on_road_non_gasoline_mileage_and_emissions

    @staticmethod
    def calculate_co2_emissions(fuel_type, fuel_usage):
        """ Calculate CO2 emissions for a mobile source fuel """
        ret = mobile_combustion_co2_emission_factors[fuel_type] * fuel_usage
        return ret

    @staticmethod
    def calculate_ch4_emissions(vehicle_type, fuel_type, usage, vehicle_year):
        """ Calculate CH4 emissions for a mobile source vehicle. usage is either fuel_usage or miles_traveled depending
        on if the vehicle_type is non-road or on road, respectively. """
        ret = mobile_combustion_ch4_and_n2o_emission_factors[vehicle_type][fuel_type][vehicle_year]['ch4_factor'] * \
              usage
        return ret

    @staticmethod
    def calculate_n2o_emissions(vehicle_type, fuel_type, usage, vehicle_year):
        """ Calculate CH4 emissions for a mobile source vehicle. usage is either fuel_usage or miles_traveled depending
        on if the vehicle_type is non-road or on road, respectively. """
        ret = mobile_combustion_ch4_and_n2o_emission_factors[vehicle_type][fuel_type][vehicle_year]['n2o_factor'] * \
              usage
        return ret

    def make_total_co2_equivalent_emissions(self):
        """ Calculate total CO2 equivalent emissions """
        total = (self._total_emissions['CO2'] + refrigerants_gwp_factors['ch4'] * self._total_emissions['CH4'] / 1000. +
                 refrigerants_gwp_factors['n2o'] * self._total_emissions['N2O'] / 1000.) / 1000.

        self._output['totalCO2EquivalentEmissions'] = total

    def make_total_biomass_co2_equivalent_emissions(self):
        """ Calculate total biomass CO2 equivalent emissions """
        total = (self.total_fuel_usage_and_co2_emissions['ethanol']['fuelUsage'] * self.ethanol_percent / 100. *
                 mobile_combustion_co2_emission_factors['ethanol'] +
                 self.total_fuel_usage_and_co2_emissions['biodiesel']['fuelUsage'] * self.biodiesel_percent / 100. *
                 mobile_combustion_co2_emission_factors['biodiesel']) / 1000.

        self._output['totalBiomassCO2EquivalentEmissions'] = total



""" Electricity models """
# pylint: disable=no-name-in-module
import logging

from atomic6ghg.formulas import Formula
from atomic6ghg.factors import electricity_emission_factors, unit_conversions_factors, refrigerants_gwp_factors

logger = logging.getLogger(__name__)


class Electricity(Formula):
    """ Calculate emissions from purchased gases """

    subregions = ['akgd', 'akms', 'aznm', 'camx', 'erct', 'frcc', 'hims',
                  'hioa', 'mroe', 'mrow', 'newe', 'nwpp', 'nycw', 'nyli',
                  'nyup', 'prms', 'rfce', 'rfcm', 'rfcw', 'rmpa', 'spno',
                  'spso', 'srmv', 'srmw', 'srso', 'srtv', 'srvc']

    def __init__(self, wks_data=None):
        super().__init__(wks_data=wks_data)
        self.recalc(self.wks_data)

    def recalc(self, wks_data: dict) -> dict:
        """Execute recalc procedure for Electricity"""
        self.wks_data = wks_data

        self.make_emissions()
        self.make_total_emissions_for_all_sources()
        self.make_co2_equivalent_emissions_location_based()
        self.make_co2_equivalent_emissions_market_based()

        return self.to_dict()

    def make_emissions(self):
        """Calculate emissions for all fuels burned"""
        self._total_emissions = {'electricityPurchased': 0.,
                                 'marketBasedEmissionsCO2Emissions': 0.,
                                 'marketBasedEmissionsCH4Emissions': 0.,
                                 'marketBasedEmissionsN2OEmissions': 0.,
                                 'locationBasedEmissionsCO2Emissions': 0.,
                                 'locationBasedEmissionsCH4Emissions': 0.,
                                 'locationBasedEmissionsN2OEmissions': 0.}
        emissions = []

        for row in self.wks_data.get('totalElectricityPurchased', []):
            subregion = row['eGridSubregion']
            source_id = row.get('sourceId')
            if not subregion:
                emissions.append(row)
                continue

            electricity_purchased = row['electricityPurchased']
            market_based_emissions_factor_co2 = row["marketBasedEmissionFactorsCO2Emissions"]
            market_based_emissions_factor_ch4 = row["marketBasedEmissionFactorsCH4Emissions"]
            market_based_emissions_factor_n2o = row["marketBasedEmissionFactorsN2OEmissions"]

            market_based_emissions_co2_emissions = \
                Electricity.calculate_market_based_emissions_electricity(electricity_purchased,
                                                                         market_based_emissions_factor_co2,
                                                                         subregion,
                                                                         'co2')
            market_based_emissions_ch4_emissions = \
                Electricity.calculate_market_based_emissions_electricity(electricity_purchased,
                                                                         market_based_emissions_factor_ch4,
                                                                         subregion,
                                                                         'ch4')
            market_based_emissions_n2o_emissions = \
                Electricity.calculate_market_based_emissions_electricity(electricity_purchased,
                                                                         market_based_emissions_factor_n2o,
                                                                         subregion,
                                                                         'n2o')
            location_based_emissions_co2_emissions = \
                Electricity.calculate_location_based_emissions_electricity(electricity_purchased,
                                                                           subregion,
                                                                           'co2')
            location_based_emissions_ch4_emissions = \
                Electricity.calculate_location_based_emissions_electricity(electricity_purchased,
                                                                           subregion,
                                                                           'ch4')
            location_based_emissions_n2o_emissions = \
                Electricity.calculate_location_based_emissions_electricity(electricity_purchased,
                                                                           subregion,
                                                                           'n2o')

            self._total_emissions['electricityPurchased'] += electricity_purchased
            self._total_emissions['marketBasedEmissionsCO2Emissions'] += market_based_emissions_co2_emissions
            self._total_emissions['marketBasedEmissionsCH4Emissions'] += market_based_emissions_ch4_emissions
            self._total_emissions['marketBasedEmissionsN2OEmissions'] += market_based_emissions_n2o_emissions
            self._total_emissions['locationBasedEmissionsCO2Emissions'] += location_based_emissions_co2_emissions
            self._total_emissions['locationBasedEmissionsCH4Emissions'] += location_based_emissions_ch4_emissions
            self._total_emissions['locationBasedEmissionsN2OEmissions'] += location_based_emissions_n2o_emissions

            emissions.append({'sourceId': source_id, 'sourceDescription': row.get('sourceDescription'),
                              'sourceArea': row.get('sourceArea'), 'eGridSubregion': subregion,
                              'electricityPurchased': row['electricityPurchased'],
                              'marketBasedEmissionFactorsCO2Emissions': row['marketBasedEmissionFactorsCO2Emissions'],
                              'marketBasedEmissionFactorsCH4Emissions': row['marketBasedEmissionFactorsCH4Emissions'],
                              'marketBasedEmissionFactorsN2OEmissions': row['marketBasedEmissionFactorsN2OEmissions'],
                              'marketBasedEmissionsCO2Emissions': market_based_emissions_co2_emissions,
                              'marketBasedEmissionsCH4Emissions': market_based_emissions_ch4_emissions,
                              'marketBasedEmissionsN2OEmissions': market_based_emissions_n2o_emissions,
                              'locationBasedEmissionsCO2Emissions': location_based_emissions_co2_emissions,
                              'locationBasedEmissionsCH4Emissions': location_based_emissions_ch4_emissions,
                              'locationBasedEmissionsN2OEmissions': location_based_emissions_n2o_emissions})

        self._output['totalElectricityPurchased'] = emissions

    def make_total_emissions_for_all_sources(self):
        """ Make total row """
        total_emissions_for_all_sources = {
            'electricityPurchased': self._total_emissions['electricityPurchased'],
            'marketBasedEmissionsCO2Emissions': self._total_emissions['marketBasedEmissionsCO2Emissions'],
            'marketBasedEmissionsCH4Emissions': self._total_emissions['marketBasedEmissionsCH4Emissions'],
            'marketBasedEmissionsN2OEmissions': self._total_emissions['marketBasedEmissionsN2OEmissions'],
            'locationBasedEmissionsCO2Emissions': self._total_emissions['locationBasedEmissionsCO2Emissions'],
            'locationBasedEmissionsCH4Emissions': self._total_emissions['locationBasedEmissionsCH4Emissions'],
            'locationBasedEmissionsN2OEmissions': self._total_emissions['locationBasedEmissionsN2OEmissions']}

        self._output['totalEmissionsForAllSources'] = total_emissions_for_all_sources

    @staticmethod
    def calculate_market_based_emissions_electricity(purchased_amount, emissions_factor, subregion, fuel):
        """Calculate CO2 emissions for a gas given material balance inputs"""
        if emissions_factor:
            ret = (purchased_amount / 1000) * emissions_factor
        else:
            ret = Electricity.calculate_location_based_emissions_electricity(purchased_amount, subregion, fuel)
        # Handle potential negative values
        ret = max(0., ret)
        return ret

    @staticmethod
    def calculate_location_based_emissions_electricity(purchased_amount, subregion, fuel):
        """Calculate CO2 emissions for a gas given material balance inputs"""
        # print(subregion, electricity_emission_factors[subregion])
        if subregion:
            ret = (purchased_amount/1000) * (electricity_emission_factors[subregion][fuel])
        else:
            ret = 0
        # Handle potential negative values
        ret = max(0., ret)
        return ret

    def make_co2_equivalent_emissions_location_based(self):
        """Calculate co2 equivalent emissions """
        location_based_emissions_co2_emissions = self._total_emissions['locationBasedEmissionsCO2Emissions']
        location_based_emissions_ch4_emissions = self._total_emissions['locationBasedEmissionsCH4Emissions']
        location_based_emissions_n2o_emissions = self._total_emissions['locationBasedEmissionsN2OEmissions']
        total = ((location_based_emissions_co2_emissions +
                  (location_based_emissions_ch4_emissions * refrigerants_gwp_factors['ch4']) +
                  (location_based_emissions_n2o_emissions * refrigerants_gwp_factors['n2o'])) *
                 unit_conversions_factors['pounds']['kilogram']) / 1000

        self._output['CO2EquivalentEmissionsLocationBasedElectricityEmissions'] = total

    def make_co2_equivalent_emissions_market_based(self):
        """Calculate co2 equivalent emissions """
        market_based_emissions_co2_emissions = self._total_emissions['marketBasedEmissionsCO2Emissions']
        market_based_emissions_ch4_emissions = self._total_emissions['marketBasedEmissionsCH4Emissions']
        market_based_emissions_n2o_emissions = self._total_emissions['marketBasedEmissionsN2OEmissions']
        total = ((market_based_emissions_co2_emissions +
                  (market_based_emissions_ch4_emissions * refrigerants_gwp_factors['ch4']) +
                  (market_based_emissions_n2o_emissions * refrigerants_gwp_factors['n2o'])) *
                 unit_conversions_factors['pounds']['kilogram']) / 1000

        self._output['CO2EquivalentEmissionsMarketBasedElectricityEmissions'] = total



""" Steam models """
# pylint: disable=no-name-in-module
import logging

from atomic6ghg.formulas import Formula
from atomic6ghg.factors import stationary_combustion_emission_factors, refrigerants_gwp_factors

logger = logging.getLogger(__name__)


class Steam(Formula):
    """ Calculate emissions from purchased gases """

    fuel_types = ['anthraciteCoal', 'bituminousCoal', 'coalCoke', 'distillateFuelOilNo2', 'kerosene', 'landfillGas',
                  'ligniteCoal', 'liquefiedPetroleumGases', 'mixedElectricPowerSector', 'naturalGas',
                  'residualFuelOilNo6', 'subBituminousCoal', 'woodAndWoodResiduals']
    default_boiler_efficiency_frac = 0.8

    def __init__(self, wks_data=None):
        super().__init__(wks_data=wks_data)
        self._total_emissions_by_fuel_type = {}
        self.recalc(self.wks_data)

    def recalc(self, wks_data: dict) -> dict:
        """Execute recalc procedure for Steam"""
        self.wks_data = wks_data

        self.make_emission_factor_data_for_steam_purchased()
        self.make_emissions_by_source_and_fuel_type()
        self.make_co2_equivalent_emissions_location_based()
        self.make_co2_equivalent_emissions_market_based()

        return self.to_dict()

    # pylint: disable=too-many-locals
    def make_emission_factor_data_for_steam_purchased(self):
        """Calculate emissions for each user input fuel source (each row)"""
        self._total_emissions_by_fuel_type = {fuel_type: {'locationBasedCO2Emissions': 0.,
                                                          'locationBasedCH4Emissions': 0.,
                                                          'locationBasedN2OEmissions': 0.,
                                                          'marketBasedCO2Emissions': 0.,
                                                          'marketBasedCH4Emissions': 0.,
                                                          'marketBasedN2OEmissions': 0.}
                                              for fuel_type in self.fuel_types}
        emission_factor_data_for_steam_purchased = []

        for row in self.wks_data.get('emissionFactorDataForSteamPurchased', []):
            fuel_type = row['fuelType']
            steam_purchased = row['steamPurchased']
            if not fuel_type or not steam_purchased:
                emission_factor_data_for_steam_purchased.append(row)
                continue
            boiler_efficiency = row['boilerEfficiency']
            if boiler_efficiency:
                boiler_efficiency_frac = boiler_efficiency / 100.
            else:
                boiler_efficiency_frac = self.default_boiler_efficiency_frac

            location_based_emission_factors_co2 = row['locationBasedEmissionFactorsCO2Factor']
            location_based_emission_factors_ch4 = row['locationBasedEmissionFactorsCH4Factor']
            location_based_emission_factors_n2o = row['locationBasedEmissionFactorsN2OFactor']
            market_based_emission_factors_co2 = row['marketBasedEmissionFactorsCO2Factor']
            market_based_emission_factors_ch4 = row['marketBasedEmissionFactorsCH4Factor']
            market_based_emission_factors_n2o = row['marketBasedEmissionFactorsN2OFactor']

            location_based_emissions_co2_emissions = \
                Steam.calculate_location_based_emissions_steam(steam_purchased, fuel_type,
                    location_based_emission_factors_co2, boiler_efficiency_frac, 'co2')

            location_based_emissions_ch4_emissions = \
                Steam.calculate_location_based_emissions_steam(steam_purchased, fuel_type,
                    location_based_emission_factors_ch4, boiler_efficiency_frac, 'ch4')

            location_based_emissions_n2o_emissions = \
                Steam.calculate_location_based_emissions_steam(steam_purchased, fuel_type,
                    location_based_emission_factors_n2o, boiler_efficiency_frac, 'n2o')

            market_based_emissions_co2_emissions = \
                Steam.calculate_market_based_emissions_steam(steam_purchased, fuel_type,
                    location_based_emission_factors_co2, market_based_emission_factors_co2, boiler_efficiency_frac,
                    'co2')

            market_based_emissions_ch4_emissions = \
                Steam.calculate_market_based_emissions_steam(steam_purchased, fuel_type,
                    location_based_emission_factors_ch4, market_based_emission_factors_ch4, boiler_efficiency_frac,
                    'ch4')

            market_based_emissions_n2o_emissions = \
                Steam.calculate_market_based_emissions_steam(steam_purchased, fuel_type,
                    location_based_emission_factors_n2o, market_based_emission_factors_n2o, boiler_efficiency_frac,
                    'n2o')

            self._total_emissions_by_fuel_type[fuel_type]['locationBasedCO2Emissions'] += \
                location_based_emissions_co2_emissions
            self._total_emissions_by_fuel_type[fuel_type]['locationBasedCH4Emissions'] += \
                location_based_emissions_ch4_emissions
            self._total_emissions_by_fuel_type[fuel_type]['locationBasedN2OEmissions'] += \
                location_based_emissions_n2o_emissions
            self._total_emissions_by_fuel_type[fuel_type]['marketBasedCO2Emissions'] += \
                market_based_emissions_co2_emissions
            self._total_emissions_by_fuel_type[fuel_type]['marketBasedCH4Emissions'] += \
                market_based_emissions_ch4_emissions
            self._total_emissions_by_fuel_type[fuel_type]['marketBasedN2OEmissions'] += \
                market_based_emissions_n2o_emissions

            calculated_row = {'sourceId': row['sourceId'], 'sourceDescription': row['sourceDescription'],
                              'sourceArea': row['sourceArea'], 'fuelType': fuel_type,
                              'boilerEfficiency': boiler_efficiency, 'steamPurchased': steam_purchased,
                              'locationBasedEmissionFactorsCO2Factor': location_based_emission_factors_co2,
                              'locationBasedEmissionFactorsCH4Factor': location_based_emission_factors_ch4,
                              'locationBasedEmissionFactorsN2OFactor': location_based_emission_factors_n2o,
                              'locationBasedEmissionsCO2Emissions': location_based_emissions_co2_emissions,
                              'locationBasedEmissionsCH4Emissions': location_based_emissions_ch4_emissions,
                              'locationBasedEmissionsN2OEmissions': location_based_emissions_n2o_emissions,
                              'marketBasedEmissionFactorsCO2Factor': market_based_emission_factors_co2,
                              'marketBasedEmissionFactorsCH4Factor': market_based_emission_factors_ch4,
                              'marketBasedEmissionFactorsN2OFactor': market_based_emission_factors_n2o,
                              'marketBasedEmissionsCO2Emissions': market_based_emissions_co2_emissions,
                              'marketBasedEmissionsCH4Emissions': market_based_emissions_ch4_emissions,
                              'marketBasedEmissionsN2OEmissions': market_based_emissions_n2o_emissions}
            emission_factor_data_for_steam_purchased.append(calculated_row)

        self._output['emissionFactorDataForSteamPurchased'] = emission_factor_data_for_steam_purchased

    @staticmethod
    # pylint: disable=too-many-arguments
    def calculate_market_based_emissions_steam(purchased_amount, fuel_type, location_based_factor, market_based_factor,
                                               boiler_efficiency_frac, gas_emission):
        """ Calculate gas (CO2, CH4, N2O) emissions for a fuel using market based emissions factors """
        if market_based_factor:
            ret = purchased_amount * market_based_factor / boiler_efficiency_frac
            # Handle potential negative values
            ret = max(0., ret)
        else:
            ret = Steam.calculate_location_based_emissions_steam(purchased_amount, fuel_type, location_based_factor,
                boiler_efficiency_frac, gas_emission)

        return ret

    @staticmethod
    def calculate_location_based_emissions_steam(purchased_amount, fuel_type, location_based_factor,
                                                 boiler_efficiency_frac, gas_emission):
        """ Calculate gas (CO2, CH4, N2O) emissions for a fuel using location based emissions factors """
        gas_emission_map = {'co2': 'CO2 Factor (kg / mmBtu)', 'ch4': 'CH4 Factor (g / mmBtu)',
                            'n2o': 'N2O Factor (g / mmBtu)'}

        if not location_based_factor:
            location_based_factor = stationary_combustion_emission_factors[fuel_type][gas_emission_map[gas_emission]]

        ret = purchased_amount * location_based_factor / boiler_efficiency_frac

        # Handle potential negative values
        ret = max(0., ret)
        return ret

    def make_emissions_by_source_and_fuel_type(self):
        """ Construct emissions_by_source_and_fuel_type table """
        self._total_emissions = {'locationBasedCO2Emissions': 0., 'locationBasedCH4Emissions': 0.,
                                 'locationBasedN2OEmissions': 0., 'marketBasedCO2Emissions': 0.,
                                 'marketBasedCH4Emissions': 0., 'marketBasedN2OEmissions': 0.}
        emissions_by_source_and_fuel_type = []
        for fuel_type in self._total_emissions_by_fuel_type:
            for k in self._total_emissions_by_fuel_type[fuel_type]:
                self._total_emissions[k] += self._total_emissions_by_fuel_type[fuel_type][k]
            row = {'fuelType': fuel_type, **self._total_emissions_by_fuel_type[fuel_type]}
            emissions_by_source_and_fuel_type.append(row)
        total_row = {'fuelType': 'totalEmissions', **self._total_emissions}
        emissions_by_source_and_fuel_type.append(total_row)

        self._output['emissionsBySourceAndFuelType'] = emissions_by_source_and_fuel_type

    def make_co2_equivalent_emissions_location_based(self):
        """ Calculate location based co2 equivalent emissions """
        location_based_emissions_co2_emissions = self._total_emissions['locationBasedCO2Emissions']
        location_based_emissions_ch4_emissions = self._total_emissions['locationBasedCH4Emissions']
        location_based_emissions_n2o_emissions = self._total_emissions['locationBasedN2OEmissions']
        total = (location_based_emissions_co2_emissions +
                 (location_based_emissions_ch4_emissions * refrigerants_gwp_factors['ch4'] / 1000) +
                 (location_based_emissions_n2o_emissions * refrigerants_gwp_factors['n2o'] / 1000)) / 1000

        self._output['CO2EquivalentEmissionsLocationBasedElectricityEmissions'] = total

    def make_co2_equivalent_emissions_market_based(self):
        """ Calculate market based co2 equivalent emissions """
        market_based_emissions_co2_emissions = self._total_emissions['marketBasedCO2Emissions']
        market_based_emissions_ch4_emissions = self._total_emissions['marketBasedCH4Emissions']
        market_based_emissions_n2o_emissions = self._total_emissions['marketBasedN2OEmissions']
        total = (market_based_emissions_co2_emissions +
                 (market_based_emissions_ch4_emissions * refrigerants_gwp_factors['ch4'] / 1000) +
                 (market_based_emissions_n2o_emissions * refrigerants_gwp_factors['n2o'] / 1000)) / 1000

        self._output['CO2EquivalentEmissionsMarketBasedElectricityEmissions'] = total



""" Base class for all formula classes """
import json
import logging
from io import StringIO

logger = logging.getLogger(__name__)


class Formula:
    """ Base class for all formula classes """
    def __init__(self, wks_data=None):
        self.wks_data = wks_data or {}

        # This variable exists for convenience of calculating totals in child classes
        self._total_emissions = {}

        self._output = {}

    def recalc(self, wks_data):
        """ All child classes must implement this method """
        raise NotImplementedError

    def to_dict(self):
        """ API to expose _output """
        try:
            # Check to ensure that _output is JSON serializable, otherwise TypeError is thrown
            io = StringIO()
            json.dump(self._output, io)
            ret = self._output
        except TypeError as e:
            err_msg = f'{self.__class__.__name__}._output is not a dict. Error message: {e}'
            logger.error(err_msg)
            ret = None
        return ret

    def to_json(self):
        """ API to expose _output as JSON """
        try:
            ret = json.dumps(self._output, indent=2)
        except TypeError as e:
            err_msg = f'{self.__class__.__name__}._output is not a dict. Error message: {e}'
            logger.error(err_msg)
            ret = None
        return ret


def null_replacer(value):
    """ Replaces None with 0. for user input float values to be used in equations """
    if not value:
        value = 0.
    return value



""" init """
# pylint: disable=no-name-in-module
from atomic6ghg.formulas.formula import Formula, null_replacer
from atomic6ghg.formulas.stationary_combustion import StationaryCombustion
from atomic6ghg.formulas.waste_gases import WasteGases
from atomic6ghg.formulas.electricity import Electricity
from atomic6ghg.formulas.purchased_offsets import PurchasedOffsets
from atomic6ghg.formulas.steam import Steam
from atomic6ghg.formulas.refrigeration_and_ac import RefrigerationAndAc
from atomic6ghg.formulas.fire_suppression import FireSuppression
from atomic6ghg.formulas.commuting import Commuting
from atomic6ghg.formulas.business_travel import BusinessTravel
from atomic6ghg.formulas.purchased_gases import PurchasedGases
from atomic6ghg.formulas.product_transport import ProductTransport
from atomic6ghg.formulas.waste import Waste
from atomic6ghg.formulas.mobile_sources import MobileSources



""" Refrigeration and AC models """
# pylint: disable=no-name-in-module
import logging

from atomic6ghg.formulas import Formula, null_replacer
from atomic6ghg.factors import refrigerants_gwp_factors, refrigeration_and_ac_equipment_emission_factors, \
    unit_conversions_factors

logger = logging.getLogger(__name__)


class RefrigerationAndAc(Formula):
    """ Calculate emissions from refrigeration and air conditioning equipment """

    def __init__(self, wks_data=None):
        super().__init__(wks_data=wks_data)
        self.recalc(self.wks_data)

    def recalc(self, wks_data: dict) -> dict:
        """Execute recalc procedure for RefrigerationAndAc"""
        self.wks_data = wks_data

        self.make_material_balance()
        self.make_simplified_material_balance()
        self.make_screening_method()
        self.make_co2_equivalent_emissions()

        return self.to_dict()

    def make_material_balance(self):
        """Calculate CO2 equivalent emissions for each input row"""
        material_balance = []
        self._total_emissions['materialBalance'] = 0.
        for row in self.wks_data.get('materialBalance', []):
            gas = row['gas']
            if not gas:
                material_balance.append(row)
                continue

            gwp = refrigerants_gwp_factors[gas]

            inventory_change = null_replacer(row['inventoryChange'])
            transferred_amount = null_replacer(row['transferredAmount'])
            capacity_change = null_replacer(row['capacityChange'])

            co2_equivalent_emissions = RefrigerationAndAc.calculate_co2_emissions_material_balance(gwp,
                inventory_change, transferred_amount, capacity_change)

            self._total_emissions['materialBalance'] += co2_equivalent_emissions

            calculated_row = {'gas': gas, 'gasGWP': gwp, 'inventoryChange': inventory_change,
                              'transferredAmount': transferred_amount, 'capacityChange': capacity_change,
                              'CO2EquivalentEmissions': co2_equivalent_emissions}
            material_balance.append(calculated_row)

        self._output['materialBalance'] = material_balance

    def make_simplified_material_balance(self):
        """Calculate CO2 equivalent emissions for each input row"""
        simplified_material_balance = []
        self._total_emissions['simplifiedMaterialBalance'] = 0.
        for row in self.wks_data.get('simplifiedMaterialBalance', []):
            gas = row['gas']
            if not gas:
                simplified_material_balance.append(row)
                continue

            gwp = refrigerants_gwp_factors[gas]

            new_units_charge = null_replacer(row['newUnitsCharge'])
            new_units_capacity = null_replacer(row['newUnitsCapacity'])
            existing_units_recharge = null_replacer(row['existingUnitsRecharge'])
            disposed_units_capacity = null_replacer(row['disposedUnitsCapacity'])
            disposed_units_recovered = null_replacer(row['disposedUnitsRecovered'])

            co2_equivalent_emissions = \
                RefrigerationAndAc.calculate_co2_emissions_simplified_material_balance(gwp, new_units_charge,
                    new_units_capacity, existing_units_recharge, disposed_units_capacity, disposed_units_recovered)

            self._total_emissions['simplifiedMaterialBalance'] += co2_equivalent_emissions

            calculated_row = {'gas': gas, 'gasGWP': gwp, 'newUnitsCharge': new_units_charge,
                              'newUnitsCapacity': new_units_capacity, 'existingUnitsRecharge': existing_units_recharge,
                              'disposedUnitsCapacity': disposed_units_capacity,
                              'disposedUnitsRecovered': disposed_units_recovered,
                              'CO2EquivalentEmissions': co2_equivalent_emissions}
            simplified_material_balance.append(calculated_row)

        self._output['simplifiedMaterialBalance'] = simplified_material_balance

    def make_screening_method(self):
        """Calculate CO2 equivalent emissions for each input row"""
        screening_method = []
        self._total_emissions['screeningMethod'] = 0.
        for row in self.wks_data.get('screeningMethod', []):
            gas = row['gas']
            type_of_equipment = row['typeOfEquipment']
            if not gas or not type_of_equipment:
                screening_method.append(row)
                continue

            gwp = refrigerants_gwp_factors[gas]

            new_units_charge = null_replacer(row['newUnitsCharge'])
            capacity_operating_units = null_replacer(row['capacityOperatingUnits'])
            capacity_disposed_units = null_replacer(row['capacityDisposedUnits'])

            co2_equivalent_emissions = RefrigerationAndAc.calculate_co2_emissions_screening_method(gwp,
                type_of_equipment, new_units_charge, capacity_operating_units, capacity_disposed_units)

            self._total_emissions['screeningMethod'] += co2_equivalent_emissions

            calculated_row = {'sourceId': row['sourceId'], 'typeOfEquipment': type_of_equipment,
                              'gas': gas, 'gasGWP': gwp, 'newUnitsCharge': new_units_charge,
                              'capacityOperatingUnits': capacity_operating_units,
                              'capacityDisposedUnits': capacity_disposed_units,
                              'CO2EquivalentEmissions': co2_equivalent_emissions}
            screening_method.append(calculated_row)

        self._output['screeningMethod'] = screening_method

    @staticmethod
    def calculate_co2_emissions_material_balance(gwp, inventory_change, transferred_amount, capacity_change):
        """Calculate CO2 emissions for a gas given material balance inputs"""
        ret = gwp * (inventory_change + transferred_amount + capacity_change)
        # Handle potential negative values
        ret = max(0., ret)
        return ret

    @staticmethod
    # pylint: disable=too-many-arguments
    def calculate_co2_emissions_simplified_material_balance(gwp, new_units_charge, new_units_capacity,
                                                            existing_units_recharge, disposed_units_capacity,
                                                            disposed_units_recovered):
        """Calculate CO2 emissions for a gas given simplified material balance inputs"""
        ret = gwp * (new_units_charge - new_units_capacity + existing_units_recharge +
                     disposed_units_capacity - disposed_units_recovered)
        # Handle potential negative values
        ret = max(0., ret)
        return ret

    @staticmethod
    def calculate_co2_emissions_screening_method(gwp, type_of_equipment, new_units_charge, capacity_operating_units,
                                                 capacity_disposed_units):
        """Calculate CO2 emissions for a gas given simplified material balance inputs"""
        installation_emission_factor = \
            refrigeration_and_ac_equipment_emission_factors[type_of_equipment]['installationEmissionFactor']
        operating_emission_factor = \
            refrigeration_and_ac_equipment_emission_factors[type_of_equipment]['operatingEmissionFactor']
        refrigerant_remaining_at_disposal = \
            refrigeration_and_ac_equipment_emission_factors[type_of_equipment]['refrigerantRemainingAtDisposal']
        recovery_efficiency = \
            refrigeration_and_ac_equipment_emission_factors[type_of_equipment]['recoveryEfficiency']

        ret = gwp * (new_units_charge * installation_emission_factor +
                     capacity_operating_units * operating_emission_factor +
                     capacity_disposed_units * refrigerant_remaining_at_disposal * (1 - recovery_efficiency))
        # Handle potential negative values
        ret = max(0., ret)
        return ret

    def make_co2_equivalent_emissions(self):
        """Calculate co2 equivalent emissions """
        total = (unit_conversions_factors['pounds']['kilogram'] * (self._total_emissions['materialBalance'] +
                  self._total_emissions['simplifiedMaterialBalance']) + self._total_emissions['screeningMethod']) / 1000.

        self._output['totalCO2EquivalentEmissions'] = total



""" Business Travel Model """
# pylint: disable=no-name-in-module
import logging

from atomic6ghg.formulas import Formula, null_replacer
from atomic6ghg.factors import business_travel_factors, refrigerants_gwp_factors

logger = logging.getLogger(__name__)


class BusinessTravel(Formula):
    """ Calculate emissions from business travel """
    personal_vehicles = ['passengerCars', 'lightDutyTruck', 'motorcycle']
    rail_or_bus = ['intercityRailNortheastCorridor', 'intercityRailOtherRoutes', 'intercityRailNationalAverage',
                      'commuterRail', 'transitRail', 'bus']
    flight_length = ['shortHaul', 'mediumHaul', 'longHaul']
    vehicle_types_all = personal_vehicles + rail_or_bus + flight_length

    def __init__(self, wks_data=None):
        super().__init__(wks_data=wks_data)

        self._total_emissions_by_vehicle = {}
        self._total_emissions_by_rail_or_bus = {}
        self._total_emissions_by_air = {}

        self.recalc(self.wks_data)

    def recalc(self, wks_data: dict) -> dict:
        """Execute recalc procedure for Commuting"""
        self.wks_data = wks_data

        self.make_personal_vehicle_rental_car_or_taxi()
        self.make_personal_vehicle_total()
        self.make_rail_or_bus()
        self.make_rail_or_bus_total()
        self.make_air_business_travel()
        self.make_air_total()
        self.make_co2_emissions_by_travel_type()
        self.make_co2_equivalent_emissions()

        return self.to_dict()

    def make_personal_vehicle_rental_car_or_taxi(self):
        """Calculate CO2 equivalent emissions for each input row in Personal Vehicle, Rental Car or Taxi table"""
        personal_vehicle = []
        self._total_emissions = {
            'personalVehicleRentalCarOrTaxiBusinessTravel': {
                'CO2': 0.,
                'CH4': 0.,
                'N2O': 0.
            }
        }
        self._total_emissions_by_vehicle = {vehicle: {
            'CO2': 0.,
            'CH4': 0.,
            'N2O': 0.
        } for vehicle in self.personal_vehicles}

        for row in self.wks_data.get('personalVehicleRentalCarOrTaxiBusinessTravel', []):
            vehicle = row['vehicleType']
            if not vehicle:
                personal_vehicle.append(row)
                continue
            if vehicle:
                travel_factor_co2 = business_travel_factors[vehicle]['co2']
                travel_factor_ch4 = business_travel_factors[vehicle]['ch4']
                travel_factor_n2o = business_travel_factors[vehicle]['n2o']
            else:
                travel_factor_co2, travel_factor_ch4, travel_factor_n2o = 0., 0., 0.

            vehicle_miles = null_replacer(row['vehicleMiles'])

            co2_emissions = BusinessTravel.calculate_transportation_emissions(travel_factor_co2, vehicle_miles)
            ch4_emissions = BusinessTravel.calculate_transportation_emissions(travel_factor_ch4, vehicle_miles)
            n2o_emissions = BusinessTravel.calculate_transportation_emissions(travel_factor_n2o, vehicle_miles)

            self._total_emissions['personalVehicleRentalCarOrTaxiBusinessTravel']['CO2'] += co2_emissions
            self._total_emissions['personalVehicleRentalCarOrTaxiBusinessTravel']['CH4'] += ch4_emissions
            self._total_emissions['personalVehicleRentalCarOrTaxiBusinessTravel']['N2O'] += n2o_emissions
            self._total_emissions_by_vehicle[vehicle]['CO2'] += co2_emissions
            self._total_emissions_by_vehicle[vehicle]['CH4'] += ch4_emissions
            self._total_emissions_by_vehicle[vehicle]['N2O'] += n2o_emissions

            calculated_row = {'sourceId': row.get('sourceId'), 'sourceDescription': row.get('sourceDescription'),
                              'vehicleType': vehicle, 'vehicleMiles': vehicle_miles, 'CO2': co2_emissions,
                              'CH4': ch4_emissions, 'N2O': n2o_emissions}

            personal_vehicle.append(calculated_row)

        self._output['personalVehicleRentalCarOrTaxiBusinessTravel'] = personal_vehicle

    def make_personal_vehicle_total(self):
        """ Make total row for personal vehicles """
        personal_vehicle_total = {'CO2': self._total_emissions['personalVehicleRentalCarOrTaxiBusinessTravel']['CO2'],
                                  'CH4': self._total_emissions['personalVehicleRentalCarOrTaxiBusinessTravel']['CH4'],
                                  'N2O': self._total_emissions['personalVehicleRentalCarOrTaxiBusinessTravel']['N2O']}

        self._output['totalForAllPersonalVehicleBusinessTravel'] = personal_vehicle_total

    def make_rail_or_bus(self):
        """Calculate CO2 equivalent emissions for each input row in Rail or Bus table"""
        rail_or_bus_vehicle = []
        self._total_emissions = {
            'railOrBusBusinessTravel': {
                'CO2': 0.,
                'CH4': 0.,
                'N2O': 0.
            }
        }

        self._total_emissions_by_rail_or_bus = {rail_or_bus: {
            'CO2': 0.,
            'CH4': 0.,
            'N2O': 0.
        } for rail_or_bus in self.rail_or_bus}

        for row in self.wks_data.get('railOrBusBusinessTravel', []):
            vehicle = row['vehicleType']
            if not vehicle:
                rail_or_bus_vehicle.append(row)
                continue
            if vehicle:
                travel_factor_co2 = business_travel_factors[vehicle]['co2']
                travel_factor_ch4 = business_travel_factors[vehicle]['ch4']
                travel_factor_n2o = business_travel_factors[vehicle]['n2o']
            else:
                travel_factor_co2, travel_factor_ch4, travel_factor_n2o = 0., 0., 0.

            vehicle_miles = null_replacer(row['passengerMiles'])

            co2_emissions = BusinessTravel.calculate_transportation_emissions(travel_factor_co2, vehicle_miles)
            ch4_emissions = BusinessTravel.calculate_transportation_emissions(travel_factor_ch4, vehicle_miles)
            n2o_emissions = BusinessTravel.calculate_transportation_emissions(travel_factor_n2o, vehicle_miles)

            self._total_emissions['railOrBusBusinessTravel']['CO2'] += co2_emissions
            self._total_emissions['railOrBusBusinessTravel']['CH4'] += ch4_emissions
            self._total_emissions['railOrBusBusinessTravel']['N2O'] += n2o_emissions
            self._total_emissions_by_rail_or_bus[vehicle]['CO2'] += co2_emissions
            self._total_emissions_by_rail_or_bus[vehicle]['CH4'] += ch4_emissions
            self._total_emissions_by_rail_or_bus[vehicle]['N2O'] += n2o_emissions

            calculated_row = {'sourceId': row.get('sourceId'), 'sourceDescription': row.get('sourceDescription'),
                              'vehicleType': vehicle, 'passengerMiles': vehicle_miles, 'CO2': co2_emissions,
                              'CH4': ch4_emissions, 'N2O': n2o_emissions}

            rail_or_bus_vehicle.append(calculated_row)

        self._output['railOrBusBusinessTravel'] = rail_or_bus_vehicle

    def make_rail_or_bus_total(self):
        """ Make total row for rail or bus """
        rail_or_bus_total = {'CO2': self._total_emissions['railOrBusBusinessTravel']['CO2'],
                             'CH4': self._total_emissions['railOrBusBusinessTravel']['CH4'],
                             'N2O': self._total_emissions['railOrBusBusinessTravel']['N2O']}

        self._output['totalForAllRailAndBusBusinessTravel'] = rail_or_bus_total

    def make_air_business_travel(self):
        """Calculate CO2 equivalent emissions for each input row in Air table"""
        air_vehicle = []
        self._total_emissions = {
            'airBusinessTravel': {
                'CO2': 0.,
                'CH4': 0.,
                'N2O': 0.
            }
        }

        self._total_emissions_by_air = {flight_length: {
            'CO2': 0.,
            'CH4': 0.,
            'N2O': 0.
        } for flight_length in self.flight_length}

        for row in self.wks_data.get('airBusinessTravel', []):
            vehicle = row['flightLength']
            if not vehicle:
                air_vehicle.append(row)
                continue
            if vehicle:
                travel_factor_co2 = business_travel_factors[vehicle]['co2']
                travel_factor_ch4 = business_travel_factors[vehicle]['ch4']
                travel_factor_n2o = business_travel_factors[vehicle]['n2o']
            else:
                travel_factor_co2, travel_factor_ch4, travel_factor_n2o = 0., 0., 0.

            vehicle_miles = null_replacer(row['passengerMiles'])

            co2_emissions = BusinessTravel.calculate_transportation_emissions(travel_factor_co2, vehicle_miles)
            ch4_emissions = BusinessTravel.calculate_transportation_emissions(travel_factor_ch4, vehicle_miles)
            n2o_emissions = BusinessTravel.calculate_transportation_emissions(travel_factor_n2o, vehicle_miles)

            self._total_emissions['airBusinessTravel']['CO2'] += co2_emissions
            self._total_emissions['airBusinessTravel']['CH4'] += ch4_emissions
            self._total_emissions['airBusinessTravel']['N2O'] += n2o_emissions
            self._total_emissions_by_air[vehicle]['CO2'] += co2_emissions
            self._total_emissions_by_air[vehicle]['CH4'] += ch4_emissions
            self._total_emissions_by_air[vehicle]['N2O'] += n2o_emissions

            calculated_row = {'sourceId': row.get('sourceId'), 'sourceDescription': row.get('sourceDescription'),
                              'flightLength': vehicle, 'passengerMiles': vehicle_miles, 'CO2': co2_emissions,
                              'CH4': ch4_emissions, 'N2O': n2o_emissions}

            air_vehicle.append(calculated_row)

        self._output['airBusinessTravel'] = air_vehicle

    def make_air_total(self):
        """ Make total row for air """
        air_total = {'CO2': self._total_emissions['airBusinessTravel']['CO2'],
                     'CH4': self._total_emissions['airBusinessTravel']['CH4'],
                     'N2O': self._total_emissions['airBusinessTravel']['N2O']}

        self._output['totalForAllAirBusinessTravel'] = air_total

    def make_co2_emissions_by_travel_type(self):
        """Calculate CO2 equivalent emissions for each input row"""
        total_commuting_type_co2 = []
        self._total_emissions['total'] = {'CO2': 0., 'CH4': 0., 'N2O': 0.}
        self._total_emissions.update(self._total_emissions_by_vehicle)
        self._total_emissions.update(self._total_emissions_by_rail_or_bus)
        self._total_emissions.update(self._total_emissions_by_air)
        for transport_type in self.vehicle_types_all:
            self._total_emissions['total']['CO2'] += self._total_emissions[transport_type]['CO2']
            self._total_emissions['total']['CH4'] += self._total_emissions[transport_type]['CH4']
            self._total_emissions['total']['N2O'] += self._total_emissions[transport_type]['N2O']

            calculated_row = {'transportType': transport_type, 'CO2': self._total_emissions[transport_type]['CO2'],
                              'CH4': self._total_emissions[transport_type]['CH4'],
                              'N2O': self._total_emissions[transport_type]['N2O']}
            total_commuting_type_co2.append(calculated_row)
        self._output['totalCO2EmissionsByTravelType'] = total_commuting_type_co2

    @staticmethod
    def calculate_transportation_emissions(travel_factor, vehicle_miles):
        """Calculate emissions for a gas given travel factor and vehicle miles inputs"""
        ret = travel_factor * vehicle_miles
        # Handle potential negative values
        ret = max(0., ret)
        return ret

    def make_co2_equivalent_emissions(self):
        """Calculate co2 equivalent emissions """
        total = (self._total_emissions['total']['CO2'] +
                 (self._total_emissions['total']['CH4'] / 1000 * refrigerants_gwp_factors['ch4']) +
                 (self._total_emissions['total']['N2O'] / 1000 * refrigerants_gwp_factors['n2o'])) / 1000
        self._output['totalCO2EquivalentEmissions'] = total



""" Purchased gases models """
# pylint: disable=no-name-in-module
import logging

from atomic6ghg.formulas import Formula, null_replacer
from atomic6ghg.factors import refrigerants_gwp_factors, unit_conversions_factors

logger = logging.getLogger(__name__)


class PurchasedGases(Formula):
    """ Calculate emissions from purchased gases """

    def __init__(self, wks_data=None):
        super().__init__(wks_data=wks_data)
        self.recalc(self.wks_data)

    def recalc(self, wks_data: dict) -> dict:
        """Execute recalc procedure for PurchasedGases"""
        self.wks_data = wks_data

        self.make_purchased_gases()
        self.make_co2_equivalent_emissions()

        return self.to_dict()

    def make_purchased_gases(self):
        """Calculate CO2 equivalent emissions for each input row"""
        purchased_gases = []
        self._total_emissions['purchasedGases'] = 0.
        for row in self.wks_data.get('purchasedGases', []):
            gas = row['gas']
            if not gas:
                purchased_gases.append(row)
                continue

            gwp = refrigerants_gwp_factors[gas]

            purchased_amount = null_replacer(row['purchasedAmount'])
            co2_equivalent_emissions = PurchasedGases.calculate_co2_emissions_purchased_gases(gwp,
                                                                                              purchased_amount)

            self._total_emissions['purchasedGases'] += co2_equivalent_emissions
            calculated_row = {'gas': gas, 'gasGWP': gwp, 'purchasedAmount': purchased_amount,
                              'CO2EquivalentEmissions': co2_equivalent_emissions}
            purchased_gases.append(calculated_row)

        self._output['purchasedGases'] = purchased_gases


    @staticmethod
    def calculate_co2_emissions_purchased_gases(gwp, purchased_amount):
        """Calculate CO2 emissions for a gas given material balance inputs"""
        ret = gwp * purchased_amount
        # Handle potential negative values
        ret = max(0., ret)
        return ret


    def make_co2_equivalent_emissions(self):
        """ Calculate co2 equivalent emissions """
        total = (unit_conversions_factors['pounds']['kilogram'] * (self._total_emissions['purchasedGases'])) / 1000.

        self._output['totalCO2EquivalentEmissions'] = total



""" Waste models """
# pylint: disable=no-name-in-module
import logging

from atomic6ghg.formulas import Formula
from atomic6ghg.factors import waste_emission_factors, unit_conversions_factors

logger = logging.getLogger(__name__)


class Waste(Formula):
    """ Calculate emissions from waste materials """

    disposal_methods = ["recycled", "landfilled", "combusted", "composted", "anaerobicallyDigestedDry",
                       "anaerobicallyDigestedWet"]

    def __init__(self, wks_data=None):
        super().__init__(wks_data=wks_data)
        self.recalc(self.wks_data)

    def recalc(self, wks_data: dict) -> dict:
        """Execute recalc procedure for Waste"""
        self.wks_data = wks_data

        self.make_waste_disposal()
        self.make_total_emissions_by_disposal_method()
        self.make_co2_equivalent_emissions()

        return self.to_dict()

    def make_waste_disposal(self):
        """Calculate CO2 equivalent emissions for each input row"""
        waste_disposal = []
        self._total_emissions = {disposal_method: 0. for disposal_method in self.disposal_methods}
        for row in self.wks_data.get('wasteDisposal', []):
            waste_material = row['wasteMaterial']
            disposal_method = row['disposalMethod']
            weight = row['weight']
            unit = row['unit']
            if not waste_material or not disposal_method or not weight or not unit:
                waste_disposal.append(row)
                continue

            co2_emissions = Waste.calculate_co2_emissions(waste_material, disposal_method, weight, unit)

            self._total_emissions[disposal_method] += co2_emissions

            calculated_row = {'sourceId': row['sourceId'], 'sourceDescription': row['sourceDescription'],
                              'wasteMaterial': waste_material, 'disposalMethod': disposal_method, 'weight': weight,
                              'unit': unit, 'CO2Emissions': co2_emissions}
            waste_disposal.append(calculated_row)

        self._output['wasteDisposal'] = waste_disposal

    def make_total_emissions_by_disposal_method(self):
        """Calculate CO2 equivalent emissions for each input row"""
        total_emissions_by_disposal_method = [{'wasteMaterial': disposal_method, 'CO2': co2_emissions}
                                              for disposal_method, co2_emissions in self._total_emissions.items()]
        self._output['totalEmissionsByDisposalMethod'] = total_emissions_by_disposal_method

    @staticmethod
    def calculate_co2_emissions(waste_material, disposal_method, weight, unit):
        """Calculate CO2 emissions for a waste material given the disposal method weight and units"""
        ret = waste_emission_factors[waste_material][disposal_method] * \
              unit_conversions_factors[unit]['shortTon'] * weight * 1000
        # Handle potential negative values
        ret = max(0., ret)
        return ret

    def make_co2_equivalent_emissions(self):
        """ Calculate co2 equivalent emissions """
        total = sum(self._total_emissions.values()) / 1000

        self._output['totalCO2EquivalentEmissions'] = total



"""atomic6ghg package configuration"""
import pathlib
from setuptools import setup

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

setup(name="atomic6ghg",
      version="1.0.3",
      description="Library of formulas, conversion factors and schemas for greenhouse gas calculations",
      long_description=README,
      long_description_content_type="text/markdown",
      url="https://github.com/Glasmor/atomic6ghg",
      author_email="kkelly@hey.com",
      author="Kevin C Kelly",
      license="BSD 3-Clause License",
      classifiers=[
          "Development Status :: 5 - Production/Stable",
          "License :: OSI Approved :: BSD License",
          "Programming Language :: Python :: 3",
          "Programming Language :: Python :: 3.10",
      ],
      packages=["atomic6ghg", "atomic6ghg.factors", "atomic6ghg.formulas", "atomic6ghg.schemas"],
      package_data={
          'atomic6ghg': ['schemas/*.json',
                         'factors/source_data/*.json',
                         'formulas/*.json'],
      },
      install_requires=[]
      )



""" init """
from .heat_content_factors import HeatContentFactors
from .stationary_combustion_emission_factors import StationaryCombustionEmissionFactors
from .refrigerants_gwp_factors import RefrigerantsGwpFactors
from .refrigeration_and_ac_equipment_emission_factors import RefrigerationAndAcEquipmentEmissionFactors
from .unit_conversions_factors import UnitConversionsFactors
from .molecular_weights_factors import MolecularWeightsFactors
from .electricity_emission_factors import ElectricityFactors
from .refrigerants_gwp_factors import RefrigerantsGwpFactors
from .fire_suppression_leak_rates_factors import FireSuppressionFactors
from .business_travel_factors import BusinessTravelFactors
from .product_transport_emission_factors import ProductTransportEmissionFactors
from .waste_emission_factors import WasteEmissionFactors
from .mobile_combustion_ch4_and_n2o_emission_factors import MobileCombustionCh4AndN2oEmissionFactors
from .mobile_combustion_co2_emission_factors import MobileCombustionCo2EmissionFactors

heat_content_factors = HeatContentFactors()
stationary_combustion_emission_factors = StationaryCombustionEmissionFactors()
refrigerants_gwp_factors = RefrigerantsGwpFactors()
unit_conversions_factors = UnitConversionsFactors()
molecular_weights_factors = MolecularWeightsFactors()
electricity_emission_factors = ElectricityFactors()
refrigerants_gwp_factors = RefrigerantsGwpFactors()
refrigeration_and_ac_equipment_emission_factors = RefrigerationAndAcEquipmentEmissionFactors()
fire_suppression_leak_rates_factors = FireSuppressionFactors()
business_travel_factors = BusinessTravelFactors()
product_transport_emission_factors = ProductTransportEmissionFactors()
waste_emission_factors = WasteEmissionFactors()
mobile_combustion_ch4_and_n2o_emission_factors = MobileCombustionCh4AndN2oEmissionFactors()
mobile_combustion_co2_emission_factors = MobileCombustionCo2EmissionFactors()



"""" Purchased offsets"""
# pylint: disable=no-name-in-module
import logging

from atomic6ghg.formulas import Formula, null_replacer

logger = logging.getLogger(__name__)


class PurchasedOffsets(Formula):
    """ Calculate emissions savings from purchased offsets """

    def __init__(self, wks_data=None):
        super().__init__(wks_data=wks_data)
        self.recalc(self.wks_data)

    def recalc(self, wks_data: dict) -> dict:
        """Execute recalc procedure for purchased offsets"""
        self.wks_data = wks_data

        self.make_purchased_offsets()
        self.make_co2_equivalent_emissions()

        return self.to_dict()

    def make_purchased_offsets(self):
        """Tabulate project level offsets across all projects into the value purchasedOffsets"""
        self._total_emissions['sumOfPurchasedOffsets'] = 0.

        for project in self.wks_data['purchasedOffsets']:
            project_offset = null_replacer(project['purchasedOffset'])
            self._total_emissions['sumOfPurchasedOffsets'] += project_offset

    def make_co2_equivalent_emissions(self):
        """ Transform total purchased offsets and present as a negative value """
        self._output['totalPurchasedOffsets'] = -self._total_emissions['sumOfPurchasedOffsets']



""" Product Transport models """
# pylint: disable=no-name-in-module
import logging

from atomic6ghg.formulas import Formula
from atomic6ghg.factors import product_transport_emission_factors, refrigerants_gwp_factors

logger = logging.getLogger(__name__)


class ProductTransport(Formula):
    """ Calculate emissions from vehicles used for product transport """

    vehicle_types_miles = ['mediumAndHeavyDutyTruck', 'lightDutyTruck', 'passengerCars']
    vehicle_types_short_ton = ['mediumAndHeavyDutyTruck', 'rail', 'aircraft', 'waterborneCraft']
    vehicle_types_all = set(vehicle_types_miles + vehicle_types_short_ton)

    def __init__(self, wks_data=None):
        super().__init__(wks_data=wks_data)
        self._total_emissions_by_vehicle_type_miles = {}
        self._total_emissions_by_miles = {}
        self._total_emissions_by_vehicle_type_short_ton = {}
        self._total_emissions_by_short_ton = {}
        self.recalc(self.wks_data)

    def recalc(self, wks_data: dict) -> dict:
        """Execute recalc procedure for ProductTransport"""
        self.wks_data = wks_data

        self.make_product_transport_by_vehicle_miles()
        self.make_total_by_vehicle_miles()
        self.make_product_transport_by_short_ton_miles()
        self.make_total_by_ton_miles()
        self.make_total_emissions_by_product_transport_type()
        self.make_co2_equivalent_emissions()

        return self.to_dict()

    def make_product_transport_by_vehicle_miles(self):
        """Calculate emissions (CO2, CH4, N2O) for each input row"""
        product_transport_by_vehicle_miles = []
        self._total_emissions_by_vehicle_type_miles = {vehicle_type: {'CO2': 0., 'CH4': 0., 'N2O': 0.}
                                                       for vehicle_type in self.vehicle_types_miles}
        self._total_emissions_by_miles = {'CO2': 0., 'CH4': 0., 'N2O': 0.}
        for row in self.wks_data.get('productTransportByVehicleMiles', []):
            vehicle_type = row['vehicleType']
            vehicle_miles = row['vehicleMiles']
            if not vehicle_type or not vehicle_miles:
                product_transport_by_vehicle_miles.append(row)
                continue

            co2_emissions = ProductTransport.calculate_vehicle_emissions(vehicle_type, vehicle_miles, 'co2',
                'vehicle-mile')
            ch4_emissions = ProductTransport.calculate_vehicle_emissions(vehicle_type, vehicle_miles, 'ch4',
                'vehicle-mile')
            n2o_emissions = ProductTransport.calculate_vehicle_emissions(vehicle_type, vehicle_miles, 'n2o',
                'vehicle-mile')

            self._total_emissions_by_vehicle_type_miles[vehicle_type]['CO2'] += co2_emissions
            self._total_emissions_by_vehicle_type_miles[vehicle_type]['CH4'] += ch4_emissions
            self._total_emissions_by_vehicle_type_miles[vehicle_type]['N2O'] += n2o_emissions

            self._total_emissions_by_miles['CO2'] += co2_emissions
            self._total_emissions_by_miles['CH4'] += ch4_emissions
            self._total_emissions_by_miles['N2O'] += n2o_emissions

            calculated_row = {'sourceId': row.get('sourceId'), 'sourceDescription': row.get('sourceDescription'),
                              'vehicleType': vehicle_type, 'vehicleMiles': vehicle_miles, 'CO2': co2_emissions,
                              'CH4': ch4_emissions, 'N2O': n2o_emissions}
            product_transport_by_vehicle_miles.append(calculated_row)

        self._output['productTransportByVehicleMiles'] = product_transport_by_vehicle_miles

    def make_total_by_vehicle_miles(self):
        """ Make total row for vehicle miles """
        total_by_vehicle_miles = {'CO2': self._total_emissions_by_miles['CO2'],
                                  'CH4': self._total_emissions_by_miles['CH4'],
                                  'N2O': self._total_emissions_by_miles['N2O']}

        self._output['totalForProductTransportByVehicleMiles'] = total_by_vehicle_miles

    def make_product_transport_by_short_ton_miles(self):
        """Calculate emissions (CO2, CH4, N2O) for each input row"""
        product_transport_by_short_ton_miles = []
        self._total_emissions_by_vehicle_type_short_ton = {vehicle_type: {'CO2': 0., 'CH4': 0., 'N2O': 0.}
                                                           for vehicle_type in self.vehicle_types_short_ton}
        self._total_emissions_by_short_ton = {'CO2': 0., 'CH4': 0., 'N2O': 0.}
        for row in self.wks_data.get('productTransportByTonMiles', []):
            vehicle_type = row['vehicleType']
            short_ton_miles = row['shortTonMiles']
            if not vehicle_type or not short_ton_miles:
                product_transport_by_short_ton_miles.append(row)
                continue

            co2_emissions = ProductTransport.calculate_vehicle_emissions(vehicle_type, short_ton_miles, 'co2',
                'ton-mile')
            ch4_emissions = ProductTransport.calculate_vehicle_emissions(vehicle_type, short_ton_miles, 'ch4',
                'ton-mile')
            n2o_emissions = ProductTransport.calculate_vehicle_emissions(vehicle_type, short_ton_miles, 'n2o',
                'ton-mile')

            self._total_emissions_by_vehicle_type_short_ton[vehicle_type]['CO2'] += co2_emissions
            self._total_emissions_by_vehicle_type_short_ton[vehicle_type]['CH4'] += ch4_emissions
            self._total_emissions_by_vehicle_type_short_ton[vehicle_type]['N2O'] += n2o_emissions

            self._total_emissions_by_short_ton['CO2'] += co2_emissions
            self._total_emissions_by_short_ton['CH4'] += ch4_emissions
            self._total_emissions_by_short_ton['N2O'] += n2o_emissions

            calculated_row = {'sourceId': row.get('sourceId'), 'sourceDescription': row.get('sourceDescription'),
                              'vehicleType': vehicle_type, 'shortTonMiles': short_ton_miles, 'CO2': co2_emissions,
                              'CH4': ch4_emissions, 'N2O': n2o_emissions}
            product_transport_by_short_ton_miles.append(calculated_row)

        self._output['productTransportByTonMiles'] = product_transport_by_short_ton_miles

    def make_total_by_ton_miles(self):
        """ Make total row for ton miles """
        total_by_ton_miles = {'CO2': self._total_emissions_by_short_ton['CO2'],
                              'CH4': self._total_emissions_by_short_ton['CH4'],
                              'N2O': self._total_emissions_by_short_ton['N2O']}

        self._output['totalForAllProductTransportByTonMiles'] = total_by_ton_miles

    def make_total_emissions_by_product_transport_type(self):
        """Calculate emissions for each vehicle type"""
        total_emissions_by_product_transport_type = []
        for vehicle_type in self.vehicle_types_all:
            zero_dict = {'CO2': 0., 'CH4': 0., 'N2O': 0.}
            short_ton_emissions = self._total_emissions_by_vehicle_type_short_ton.get(vehicle_type, zero_dict)
            miles_emissions = self._total_emissions_by_vehicle_type_miles.get(vehicle_type, zero_dict)

            co2_emissions = short_ton_emissions['CO2'] + miles_emissions['CO2']
            ch4_emissions = short_ton_emissions['CH4'] + miles_emissions['CH4']
            n2o_emissions = short_ton_emissions['N2O'] + miles_emissions['N2O']

            row = {'transportType': vehicle_type, 'CO2': co2_emissions, 'CH4': ch4_emissions, 'N2O': n2o_emissions}

            total_emissions_by_product_transport_type.append(row)

        self._output['totalEmissionsByProductTransportType'] = total_emissions_by_product_transport_type

    @staticmethod
    def calculate_vehicle_emissions(vehicle_type, usage, gas_emissions, units):
        """Calculate gas emissions (CO2, CH4, N2O) emissions for a vehicle"""
        ret = product_transport_emission_factors[vehicle_type][units][gas_emissions] * usage
        # Handle potential negative values
        ret = max(0., ret)
        return ret

    def make_co2_equivalent_emissions(self):
        """ Calculate co2 equivalent emissions """
        co2_total = self._total_emissions_by_miles['CO2'] + self._total_emissions_by_short_ton['CO2']
        ch4_total = self._total_emissions_by_miles['CH4'] + self._total_emissions_by_short_ton['CH4']
        n2o_total = self._total_emissions_by_miles['N2O'] + self._total_emissions_by_short_ton['N2O']

        total = (refrigerants_gwp_factors['co2'] * co2_total +
                 refrigerants_gwp_factors['ch4'] * ch4_total / 1000 +
                 refrigerants_gwp_factors['n2o'] * n2o_total / 1000) / 1000

        self._output['totalCO2EquivalentEmissions'] = total
